extern "C" {
#  include <X11/X.h>
#  include <X11/Xatom.h>
#  include <X11/Xmd.h>
#  include <X11/Xft/Xft.h>
}
#include "Imlib2pp.hh"

#include <sys/time.h>

#include <vector>
#include <string>
#include <utility>

#include "DrawOp.hh"
#include "ImageControl.hh"
#include "ExpressionEvaluator.hh"
#include "Texture.hh"

DrawOps buttonBackground() {
    DrawOps ops;
    ImColor clr1(255, 255, 255, 0);
    ImColor clr2(255, 255, 255, 128);
    ImColor clr3(218, 178, 82);

    std::vector<std::pair<ImColor, int> > range;

    range.push_back(std::make_pair(clr2, 0));
    range.push_back(std::make_pair(clr1, 100));

    ops.push_back(new DrawOpGradient(range, "0", "0", "width", "height", -45));
    ops.push_back(new DrawOpBevel(clr2, 2, 1));
    //    ops.push_back(new DrawOpRect(clr3, "0", "0", "width", "height"));
    //    ops.push_back(new DrawOpGradient(range, "0", "0", "width", "height", -45));
    
    return ops;
}

Texture* closeButton() {
    DrawOps ops = buttonBackground();

    ImColor clr2(33, 33, 33, 255);
    
    ops.push_back(new DrawOpLine(clr2, 
                                 "5", "5", 
                                 "width-5", "height-5"));
    // /
    ops.push_back(new DrawOpLine(clr2, 
                                 "width-5","5",
                                 "5", "height-5"));

    return new Texture(ops);    
}

Texture* minimizeButton() {
    DrawOps ops = buttonBackground();

    ImColor clr2(33, 33, 33, 255);
    
    // _
    ops.push_back(new DrawOpLine(clr2, 
                                 "5", "height-8",
                                 "width-5", "height-8"));

    return new Texture(ops);    
}

Texture* titleBar() {
    DrawOps ops;
    ImColor blue(51, 102, 153, 0);
    ImColor darkBlue(0, 51, 102);
    std::vector<std::pair<ImColor, int> > gradient;
    gradient.push_back(std::make_pair(blue, 0));
    gradient.push_back(std::make_pair(darkBlue, 20));

    ops.push_back(new DrawOpFillRect(blue, "0", "0", "width", "height"));
    ops.push_back(new DrawOpGradient(gradient, 
                                     "0", "0", "width", "height", 0));
    //    ops.push_back(new DrawOpInterlace(96, 64));

    return new Texture(ops);
}

class ChildWindow {
public:
    ChildWindow(Display* _display, int _screen, Window _parent,
				unsigned int _width, unsigned int _height): 
        display(_display), screen(_screen), parent(_parent),
		width(_width), height(_height)
    {
        id = createWindow(0, 0, width, height);
		show();
    }

    virtual ~ChildWindow() {};

	bool isVisible() { return visible; }

	void hide() {
		visible = false;

		XUnmapWindow(display, id);
	}

	void show() {
		visible = true;

		XMapRaised(display, id);
	}

	void setVisibility(bool shouldBeVisible) {
		if (shouldBeVisible == visible) return;

		// true, true and false, false cases excluded
		// so only true, false (show) and false, true (hide) are possible

		if (shouldBeVisible && !visible) 
			show();
		else 
			hide();
	}

	unsigned int getWidth() { return width; }
	unsigned int getHeight() { return height; }

    void resize(unsigned int _width, unsigned int _height) {
        width  = _width;
        height = _height;

        printf("Resizing to %u and %u\n", width, height);
        XResizeWindow(display, id, width, height);
    }

    void moveTo(int _x, int _y) {
        x  = _x;
        y  = _y;

        printf("Moving to %d and %d\n", x, y);
        XMoveWindow(display, id, x, y);
    }

    virtual void repaint(ImageControl& control)=0;
protected:
    Window createWindow(int x, int y, int width, int height) {
        XSetWindowAttributes attrib_set;
        
        return XCreateWindow(display, parent, x, y, width, height, 0,
                             CopyFromParent,
                             CopyFromParent, CopyFromParent, 0,
                             &attrib_set);
    }

    
    Window id, parent;
    int screen;
    Display* display;    
    int x, y;
    unsigned int width, height;
	bool visible;
};

class Edge: public ChildWindow {
public:
	Edge(Display* _display, int _screen, Window _parent): 
		ChildWindow(_display, _screen, _parent, 10, 1)
		{
			texture = dummyTexture();
		}

	void repaint(ImageControl& control) {
		DrawOps ops;
		ops.push_back(new DrawOpTexture(texture, 0, 0, width, height));

		XSetWindowBackgroundPixmap(display, id, 
								   control.render(id, ops, 
												  0, 0, width, height));
		XClearWindow(display, id);
	}
private:
	Texture* texture;

	Texture* dummyTexture() {
		DrawOps ops;
		ImColor orange(204, 128, 0);
		
		ops.push_back(new DrawOpFillRect(orange, "0", "0", "width", "height"));
		
		return new Texture(ops);
	}
};

class Grip: public ChildWindow {
public:
	Grip(Display* _display, int _screen, Window _parent): 
		ChildWindow(_display, _screen, _parent, 20, 5)
		{
			texture = dummyTexture();
		}

	void repaint(ImageControl& control) {
		DrawOps ops;
		ops.push_back(new DrawOpTexture(texture, 0, 0, width, height));

		XSetWindowBackgroundPixmap(display, id, 
								   control.render(id, ops, 
												  0, 0, width, height));
		XClearWindow(display, id);
	}
private:
	Texture* texture;

	Texture* dummyTexture() {
		DrawOps ops;
		ImColor green(0, 128, 0);
		
		ops.push_back(new DrawOpFillRect(green, "0", "0", "width", "height"));
		
		return new Texture(ops);
	}
};

class Handle: public ChildWindow {
public:
	Handle(Display* _display, int _screen, Window _parent): 
		ChildWindow(_display, _screen, _parent, 200, 5)
		{
			texture = dummyTexture();
		}

	void repaint(ImageControl& control) {
		DrawOps ops;
		ops.push_back(new DrawOpTexture(texture, 0, 0, width, height));

		XSetWindowBackgroundPixmap(display, id, 
								   control.render(id, ops, 
												  0, 0, width, height));
		XClearWindow(display, id);
	}
private:
	Texture* texture;

	Texture* dummyTexture() {
		DrawOps ops;
		ImColor blue(0, 0, 255);
		
		ops.push_back(new DrawOpFillRect(blue, "0", "0", "width", "height"));
		
		return new Texture(ops);
	}
};

class TitleBar: public ChildWindow {
public:
    TitleBar(Display* _display, int _screen, Window _parent, 
             XftFont* font): 
        ChildWindow(_display, _screen, _parent, 1, 30)
    {
        xftdraw = XftDrawCreate(display, (Drawable) parent,
                                DefaultVisual(display, screen), 
                                DefaultColormap(display, screen));
        xftfont = font;
        texture = titleBar();
    }

    ~TitleBar() {        
        XftDrawDestroy(xftdraw);
    }

    Texture* draw() {
        return texture;
    }
    
    void drawLabel(Pixmap pixmap, int width, int height) {
        XftColor xftc, xftc_shadow;
        xftc.color.red = 0xFFFF;
        xftc.color.green = 0xFFFF;
        xftc.color.blue = 0xFFFF;
        xftc.color.alpha = 0xFFFF / 2;
        xftc.pixel = WhitePixel(display, screen);

        int x = height * 3 + 2;
        int y = height / 2 + 2 + 5;

        std::string s = "emacs@localhost";
        unsigned int len = s.length();

        XftDrawChange(xftdraw, pixmap);

        xftc_shadow.color.red = 0;
        xftc_shadow.color.green = 0;
        xftc_shadow.color.blue = 0;
        xftc_shadow.pixel = WhitePixel(display, screen);

        xftc_shadow.color.alpha = (int)(0xFFFF * 0.25);
        XftDrawStringUtf8(xftdraw, &xftc_shadow, xftfont,
                          x + 1, y + 2, (unsigned char *) s.c_str(), len);
        xftc_shadow.color.alpha = (int)(0xFFFF * 0.3);
        XftDrawStringUtf8(xftdraw, &xftc_shadow, xftfont,
                          x + 2, y + 2, (unsigned char *) s.c_str(), len);
        xftc_shadow.color.alpha = (int)(0xFFFF * 0.6);
        XftDrawStringUtf8(xftdraw, &xftc_shadow, xftfont,
                          x + 1, y + 1, (unsigned char *) s.c_str(), len);
            
        XftDrawStringUtf8(xftdraw, &xftc, 
                          xftfont, x, y,
                          (const FcChar8*)s.c_str(), s.length());
    }

    DrawOps* render() {
        Texture* texture = draw();
        DrawOps* ops = new DrawOps;

        height = 30;

        int buttonSize = height - 4;

        ops->push_back(new DrawOpTexture(titleBar(), 0, 0, width, height));
        ops->push_back(new DrawOpTexture(minimizeButton(), 
                                        0, 2, buttonSize, 
                                        buttonSize));

        ops->push_back(new DrawOpTexture(closeButton(), 
                                         width - buttonSize, 2, buttonSize, 
                                         buttonSize));
        return ops;
    }

    void repaint(ImageControl& control) {
        DrawOps* ops = render();
        Pixmap pixmap = control.render(id, *ops,
                                       0, 0, width, height);
        delete ops;
        
        drawLabel(pixmap, width, height);
        
        XSetWindowBackgroundPixmap(display, id, pixmap);
        XClearWindow(display, id);
    }
private:
    XftDraw* xftdraw;
    XftFont* xftfont;
    Texture* texture;    
};

typedef struct {
    bool title;
    bool border;
    bool handle;
    bool sticky;
    bool shaded;
    bool max;
    bool all;
    bool alwaysontop;
    bool alwaysatbottom;
    bool focusable;
    bool tasklist;
    bool fullscreen;
    bool hidden;
} KaWindowFlags;

class WindowDecorationManager {
public:
	WindowDecorationManager(Display* _display, unsigned int _screen, 
							Window _id): id(_id), display(_display),
										 screen(_screen) {
	    XftFont* xftfont2 = XftFontOpenName(display, screen, "sans-10");

		titleBar  = new TitleBar(display, screen, id, xftfont2);
		leftEdge  = new Edge(display, screen, id);
		rightEdge = new Edge(display, screen, id);

		leftGrip   = new Grip(display, screen, id);
		rightGrip  = new Grip(display, screen, id);		
		handle     = new Handle(display, screen, id);

		flags.title = flags.border = flags.handle = flags.all = true;

		oldDecorLeft = decorLeft();
		oldDecorTop  = decorTop();
		oldDecorRight = decorRight();
		oldDecorBottom  = decorBottom();
	}

	~WindowDecorationManager() {
		delete titleBar;
		delete leftEdge;
		delete rightEdge;
		delete leftGrip;
		delete rightGrip;
		delete handle;
	}

	unsigned int decorLeft() {
		unsigned int n = 0;
		if (flags.border) n += leftEdge->getWidth();
		return n;
	}

	unsigned int decorRight() {
		unsigned int n = 0;
		if (flags.border) n += rightEdge->getWidth();
		return n;
	}

	unsigned int decorTop() {
		unsigned int n = 0;
		if (flags.title) n += titleBar->getHeight();
		return n;
	}

	unsigned int decorBottom() {
		unsigned int n = 0;
		if (flags.handle) n += std::max(std::max(leftGrip->getHeight(), 
												 rightGrip->getHeight()), 
										handle->getHeight());
		return n;
	}

	void flagsChanged() {
		int dl = decorLeft() - oldDecorLeft;
		int dt = decorTop() - oldDecorTop;
		int dr = decorRight() - oldDecorRight;
		int db = decorBottom() - oldDecorBottom;
		
		printf("deltas: (%d, %d, %d, %d)\n", dl, dt, dr, db);

		// Need to adjust frame size 
		width  += dl + dr;
		height += dt + db;

		Window dummy;
		XTranslateCoordinates(display, id, RootWindow(display, screen),
							  0, 0, &x, &y, &dummy);
		printf("current coordinates: (%d, %d)\n", x, y);
		
		// And possibly move it too
		if (dl != 0 || dt != 0) {
			x += dl;
			y += dt;
			printf("moving to (%d, %d)\n", x, y);
			XMoveResizeWindow(display, id, x, y, width, height);
		} else 	if (dl + dr != 0 || dt + db != 0) {
			XResizeWindow(display, id, width, height);
		}

		titleBar->setVisibility(flags.title);

		leftEdge->setVisibility(flags.border);
		rightEdge->setVisibility(flags.border);

		leftGrip->setVisibility(flags.handle);
		rightGrip->setVisibility(flags.handle);
		handle->setVisibility(flags.handle);

		onResize(width, height);

		oldDecorLeft = decorLeft();
		oldDecorTop  = decorTop();
		oldDecorRight = decorRight();
		oldDecorBottom  = decorBottom();
	}

	void onResize(int newWidth, int newHeight) {
		width = newWidth; height = newHeight;

		titleBar->resize(width, titleBar->getHeight());

		leftEdge->moveTo(0, decorTop());
		leftEdge->resize(leftEdge->getWidth(), 
						 height - decorTop() - decorBottom());

		rightEdge->moveTo(width  - rightEdge->getWidth(), 
						  decorTop());
		rightEdge->resize(rightEdge->getWidth(), 
						  height - decorTop() - decorBottom());

		leftGrip->moveTo(0, height - leftGrip->getHeight());
		rightGrip->moveTo(width - rightGrip->getWidth(), 
						  height - rightGrip->getHeight());

		handle->moveTo(leftGrip->getWidth(), height - handle->getHeight());
		handle->resize(width - leftGrip->getWidth() - rightGrip->getWidth(),
					   handle->getHeight());
	}

	void repaintAll(ImageControl& control) {
		titleBar->repaint(control);
		leftEdge->repaint(control);
		rightEdge->repaint(control);
		leftGrip->repaint(control);
		rightGrip->repaint(control);
		handle->repaint(control);
	}

	KaWindowFlags flags;

private:
	Window id;
	ChildWindow *titleBar, *leftEdge, *rightEdge, *leftGrip, *rightGrip,
		*handle;
	int x, y;
	unsigned int width, height;
	Display* display;
	unsigned int oldDecorLeft, oldDecorTop, oldDecorRight, oldDecorBottom;
	unsigned int screen;
};

int main(int argc, char **argv) {
    Display *display = XOpenDisplay(NULL);
    int screen = DefaultScreen(display);
    Colormap colormap = DefaultColormap(display, screen);
    Visual *visual = DefaultVisual(display, screen);

    ImageControl control(display, screen, visual, colormap);

    Window win = XCreateSimpleWindow(display, RootWindow(display, screen), 100, 100, 400, 300, 0, BlackPixel(display, screen), BlackPixel(display, screen));
    XMapWindow(display, win);

    XSetWindowAttributes attr_set;
    attr_set.event_mask = StructureNotifyMask | SubstructureNotifyMask | KeyPressMask;
    XChangeWindowAttributes(display, win, CWEventMask, &attr_set);

    DrawOps ops;
    ImColor clr(255, 0, 0, 128);
    ImColor clr2(0, 255, 0, 128);
    ImColor clr3(0, 0, 255, 64);
    ImColor clr4(192, 192, 192, 128);

    DrawOps buttonOps;
    buttonOps.push_back(new DrawOpFillRect(clr4, "0", "0", "width", "height"));
    buttonOps.push_back(new DrawOpBevel(clr4, 4, 1));
    Texture buttonTexture(buttonOps);

    //    ops.push_back(new DrawOpFillRect(clr, "0", "0", 
    //                                     "width/4", "height/3"));
    //    ops.push_back(new DrawOpFillRect(clr2, 
    //                                     50, 50, 200, 200));
    //    ops.push_back(new DrawOpLine(clr3, "50", "50", "width-50", "height-50"));
    //    ops.push_back(new DrawOpInterlace(64, 96));
    //    ops.push_back(new DrawOpTexture(&buttonTexture, 0, 0, 200, 200));
    int titleHeight = 24;
    int buttonSize = (int)(titleHeight * 0.7);

    //    ops.push_back(new DrawOpTexture(titleBar(), "0", "0", "width", buf));
    //    ops.push_back(new DrawOpTexture(closeButton(), titleHeight/6, 
    //                                    titleHeight/6, buttonSize, buttonSize));
    //    ops.push_back(new DrawOpTexture(minimizeButton(), titleHeight/6 + buttonSize + 5, 
    //                                    titleHeight/6, buttonSize, buttonSize));
    //    ops.push_back(new DrawOpImage("/home/eugene/testalpha.png", 0, 0));

    std::vector<std::pair<ImColor, int> > gradient;
    gradient.push_back(std::make_pair(clr, 0));
    gradient.push_back(std::make_pair(clr2, 50));
    gradient.push_back(std::make_pair(clr3, 100));
    //    ops.push_back(new DrawOpGradient(gradient, 200, 75, 120, 120, 45));
    //    ops.push_back(new DrawOpTexture(&buttonTexture, "width/4", 
    //                                    "height/2-50", "width/2", "400"));
   

    Pixmap pixmap = control.render(win, ops, 100, 100, 400, 300);

    XSetWindowBackgroundPixmap(display, win, pixmap);
    XClearWindow(display, win);

    XEvent event;
    unsigned int w, h;

    XftDraw *xftdraw = XftDrawCreate(display, (Drawable) win,
                                     visual, colormap);

    XftFont* xftfont = XftFontOpenName(display, screen, "sans-40");
    XftFont* xftfont2 = XftFontOpenName(display, screen, "sans-10");


    XftColor xftc, xftc2;
    xftc.color.red = 0;
    xftc.color.green = 0;
    xftc.color.blue = 0;
    xftc.color.alpha = 0xFFFF / 2;
    xftc.pixel = BlackPixel(display, screen);

    xftc2.color.red = 0xFFFF;
    xftc2.color.green = 0xFFFF;
    xftc2.color.blue = 0xFFFF;
    xftc2.color.alpha = 0xFFFF / 2;
    xftc2.pixel = WhitePixel(display, screen);

	WindowDecorationManager decorManager(display, screen, win);

    while (1) {
        XNextEvent(display, &event);
        switch (event.xany.type) {
        case ConfigureNotify:
            if (event.xconfigure.window != win)
                break;

#if 0
            printf("Configure\n");
            XFreePixmap(display, pixmap);
//            titleBar.resize(event.xconfigure.width, 30);
            XSync(display, false);
#endif            
            struct timeval tv, tv2;
            gettimeofday(&tv, NULL);
#if 0            
            pixmap = control.render(win, ops, 
                                    0, 0,
                                    event.xconfigure.width, 
                                    event.xconfigure.height);

            char buf[128];
            sprintf(buf, "%d x %d", 
                    event.xconfigure.width, 
                    event.xconfigure.height);

            XftDrawChange(xftdraw, pixmap);
                       XftDrawStringUtf8(xftdraw, &xftc, xftfont, 120, 120,
                                          (const FcChar8*)buf, strlen(buf));
#endif
#if 0
            XftDrawStringUtf8(xftdraw, &xftc, xftfont2, buttonSize * 3+2, titleHeight/2+5 + 2,
                              (const FcChar8*)"emacs@localhost", 
                              strlen("emacs@localhost"));
            XftDrawStringUtf8(xftdraw, &xftc2, xftfont2, buttonSize * 3, titleHeight/2 + 5,
                              (const FcChar8*)"emacs@localhost", 
                              strlen("emacs@localhost"));
#endif 
            XSetWindowBackground(display, win, BlackPixel(display, screen));
            //XSetWindowBackgroundPixmap(display, win, pixmap);
            XClearWindow(display, win);

			decorManager.onResize(event.xconfigure.width, 
								  event.xconfigure.height);
			decorManager.repaintAll(control);

            gettimeofday(&tv2, NULL);
            XSync(display, false);


            printf("Rendered in %.4f second\n",
                   (tv2.tv_sec + tv2.tv_usec * 1e-6) - 
                   (tv.tv_sec + tv.tv_usec * 1e-6));
            break;
		case KeyPress:
			if (event.xkey.keycode == 28 /* 't' */) {
				decorManager.flags.title = !decorManager.flags.title;
				decorManager.flagsChanged();
			} else if (event.xkey.keycode == 43 /* 'h' */) {
				decorManager.flags.handle = !decorManager.flags.handle;
				decorManager.flagsChanged();
			} else if (event.xkey.keycode == 56 /* 'b' */) {
				decorManager.flags.border = !decorManager.flags.border;
				decorManager.flagsChanged();
			}
			break;
        case UnmapNotify:
			if (event.xunmap.window != win) 
				break;
            return 0;
        case ClientMessage:
            return 0;
        default:
            printf("Event received: %d\n", event.xany.type);
        }
    }
}
