# $Id: resources.py,v 1.4 2003/06/15 16:12:04 brailsmt Exp $

from kahakai import *
import re

class Stub: pass

class ResourceFile:    
    def __init__(self, filename):
        self._patterns = {}
        self._values   = {}
        self._parse(filename)

    def __getitem__(self, key):
        return self.get(key)

    def get(self, key, default=Stub):        
        if key in self._values:
            return self._values[key]
        else:
            value = self._searchPatternsByKey(key)

            if value:
                return value
            
            if default is not Stub:
                return default
            else:
                raise KeyError("Key %s not found in ResourceFile" % key)


    def _searchPatternsByKey(self, key):
        for pat, value in self._patterns.items():
            if pat.match(key):
                return value

        return None

    def _parse(self, filename):
        f = open(filename)

        for line in f:
            if not re.match(r'^\s*!', line):
                match = re.match(r'^\s*(.+?)\s*:\s*(.+)\s*$', line)
                if match:
                    pattern, value = [x.strip() for x in match.groups()]
                    if '*' not in pattern: 
                        self._values[pattern] = value
                    else:
                        regex = re.escape(pattern).replace('\*', '.*')
                        self._patterns[re.compile(regex)] = value

        f.close()


class StyleLoader:    
    def loadStyle(self, wascreen, filename, xft=True):
        self.wascreen = wascreen
        
        screen = wascreen.screen_number
        wstyle = wascreen.wstyle
        mstyle = wascreen.mstyle
        ic     = wascreen.ic

        default_font = KaFont()

        if xft:
            default_font.xft = True
            default_font.font = 'arial:pixelsize=12'
        else:
            default_font.xft = False
            default_font.font = 'fixed'

        self.rfile = ResourceFile(filename)

        height = self.readFont('window', wstyle.wa_font, default_font)
        if not wstyle.title_height: wstyle.title_height = height + 4
        
        self.readFont('menu.frame', mstyle.wa_f_font, wstyle.wa_font)
        height = self.readFont('menu.title', mstyle.wa_t_font,
                               mstyle.wa_f_font)
        if not mstyle.title_height: mstyle.title_height = height + 4

        wstyle.title_height = max(10, wstyle.title_height)
        mstyle.title_height = max(4,  mstyle.title_height)
        mstyle.item_height  = max(4, mstyle.item_height)
        
        self.readFont('menu.bullet', mstyle.wa_b_font, mstyle.wa_f_font)
        self.readFont('menu.checkbox.true',
                      mstyle.wa_ct_font, mstyle.wa_f_font)
        self.readFont('menu.checkbox.false',
                      mstyle.wa_cf_font, mstyle.wa_f_font)

        wstyle.y_pos = (wstyle.title_height / 2 - 2) + \
                       wstyle.wa_font.diff / 2 + wstyle.wa_font.diff % 2

        mstyle.f_y_pos = (mstyle.item_height / 2 - 2) + \
                       mstyle.wa_f_font.diff / 2 + mstyle.wa_f_font.diff % 2

        mstyle.f_y_pos = (mstyle.item_height / 2 - 2) + \
                         mstyle.wa_f_font.diff / 2 + mstyle.wa_f_font.diff % 2

        wstyle.wa_font_u = KaFont()
        wstyle.wa_font_u.font = wstyle.wa_font.font
        wstyle.wa_font_u.xft  = wstyle.wa_font.xft
        wstyle.wa_font_u.Open(self.wascreen.display,
                              self.wascreen.screen_number, None)
    
        #for x in 'b', 'ct', 'cf':
#            diff = getattr(mstyle, 'wa_%s_font' % x).diff 
#            setattr(mstyle, '%s_y_pos',
#                    (mstyle.item_height / 2) + diff / 2 + diff % 2)
        
        
    def readFont(self, resourceName, font, defaultFont):
        value = self.rfile.get(resourceName + '.font', None) 

        if value:
            if value.find('[xft]') != -1:
                font.xft = True
                font.font = value[:value.find('[xft]')].strip()
            else:
                font.xft  = False
                font.font = 'fixed'
        else:
            font.xft  = defaultFont.xft
            font.font = defaultFont.font

        print "-- Read font %s - %s (%s)" % (resourceName, font, font.font)

        return font.Open(self.wascreen.display,
                         self.wascreen.screen_number, None)

    
#rfile = ResourceFile('/home/eugene/.kahakai/styles/Aqua/Aqua.style')
#print rfile._values
#print rfile._patterns
#print rfile['window.title.height']
#print rfile['dockappholder..borderColor']

