from xml.sax.saxutils import quoteattr, escape

import sys

class ConfigurationWriter(object):
    def write(self, configuration):
        print '-' * 40

        self.level = 0
        self.out   = sys.stdout
        
        self._start('configuration')
        self._writeBindings(configuration.bindings)
        self._writeSettings(configuration.settings)
        self._end('configuration')
        
        print
        print '-' * 40

    def _start(self, tagName, attr=None):
        if attr is None: attr = {}

        assert len(attr.keys()) <= 1

        if attr == {}:
            print >>self.out, "%s<%s>" % (' ' * self.level, tagName)
        else:
            print >>self.out, '%s<%s %s=%s>' % (' ' * self.level,
                                                tagName, attr.keys()[0],
                                                quoteattr(attr.values()[0]))
        
        self.level += 1

    def _end(self, tagName):
        self.level -= 1
        print >>self.out, "%s</%s>" % (' ' * self.level, tagName)

    def _data(self, data):
        print >>self.out, data

    def _valueTag(self, name, value):
        if isinstance(value, list):
            for val in value:
                print >>self.out, "%s<%s>%s</%s>" % (' ' * self.level,
                                         name, escape(val), name)
        else:
            print >>self.out, "%s<%s>%s</%s>" % (' ' * self.level,
                                                 name, escape(value), name)
        
    def _writeBindings(self, bindings):
        self._start('bindings')

        for target in bindings:
            self._start('target', {'id': target})
            for eventType, params, action in bindings[target]:
                self._start('binding')
                self._start('event')
                self._start(eventType)
            
                for paramName, paramValue in params.items():
                    self._valueTag(paramName, paramValue)
                
                self._end(eventType)
                self._end('event')

                self._valueTag('action', action)
                self._end('binding')
                
            self._end('target')
        
        self._end('bindings')

    def _writeSettings(self, settings):
        self._start('settings')
        
        for name, value in settings.items():
            self._valueTag(name, value)
            
        self._end('settings')
        
