#
# tools/kaconf/kaconf.py
#
# Made by (PrinceOfChaos)
# SF Username: pervago
#

import pygtk

pygtk.require('2.0')

import gtk
import gtk.gdk as gdk
import gtk.glade
import gobject

from configuration import Configuration

from BindingDialog import BindingDialog

from settings import *

menuSettings = [
    ('Navigation', [(SimpleSetting,
                     'Mouse menu navigation', 'mouseMenuNavigation'),
                    (SimpleSetting,
                     'Keyboard menu navigation', 'keyboardMenuNavigation')]),
    ('Root menu', 
     [(ButtonSetting, 'Show by clicking desktop with', 'rootMenuClick'),
      (ButtonSetting, 'Hide by clicking desktop with', 'rootMenuUnmap'),
      ]),
    ('Window menu', 
     [(ButtonSetting, 'Show by clicking decoration with', 'windowMenuClick'),
      (ButtonSetting, 'Hide by clicking decoration with', 'windowMenuUnmap'),
      ]),    
    ]

windowSettings = [
    ('Placement', [
    (SimpleSetting,
     'Smart window _placement', 'smartWindowPlacement')]),
    ('Moving and resizing', [
    (NumericSetting, 'Window _snapping', 'windowSnapping'),
    (ModifierSetting,
     '_Move window with modifier', 'moveWindowWithModifier'),
    (ModifierSetting,
     '_Resize window with modifier', 'resizeWindowWithModifier'),
    (ModifierSetting,
     '_Lower window with modifier', 'lowerWindowWithModifier'),
    (ButtonSetting,     
     'Move by clicking on _titlebar with', 'moveByTitlebar')]),
    ('Focus and stacking',
     [(FocusSetting, '_Focus', 'focus'),
      (SimpleSetting, 'Focus new windows', 'focusNewWindows'),
      (SimpleSetting, 'Click to raise', 'clickToRaise')])
    ]

screenSettings = [
    ('Desktops',
     [(DesktopKeysSetting, 'Switch desktops with', 'desktopKeys')]
     ),
    ('Viewports',
     [(ModifierSetting, 'Surf viewports with modifier', 'viewportSurf'),
      (ModifierButtonSetting, 'Knock on edges with', 'viewportKnock')]
     )
    
    ]

class MainWindow(gtk.glade.XML):
    def __init__(self, file):
        gtk.glade.XML.__init__(self, file)
        self.signal_autoconnect(self)

        self.bindingsList = self.get_widget('bindings_list')

        renderer = gtk.CellRendererText()

        self.bindingsList.append_column(
            gtk.TreeViewColumn('Event', renderer, text=0))
        
        self.bindingsList.append_column(
            gtk.TreeViewColumn('Action', renderer, text=1))

        store = gtk.ListStore(str, str)
        
        self.bindingsList.set_model(store)

        self.fillContextsMenu()
        self.loadConfiguration()

        # Set-up initial context
        self.changeToContext(self.contexts()[0][1])

        self.settings = []
        # Add settings controls
        self.addSettingsControls(windowSettings,
                                 self.get_widget('window_settings_vbox'))
        self.addSettingsControls(menuSettings,
                                 self.get_widget('menu_settings_vbox'))
        self.addSettingsControls(screenSettings,
                                 self.get_widget('screen_settings_vbox'))
        
    def addSettingsControls(self, settings, settingsVBox):
        settingsVBox.set_spacing(18)
        
        sizeGroup = gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL)
        sizeGroup2 = gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL)

        for groupName, groupSettings in settings:
            group_vbox = gtk.VBox(spacing=6)

            label = gtk.Label('<span weight="bold">%s</span>' % \
                              groupName)
            label.set_alignment(0.0, 0.5)
            label.set_use_markup(True)
            group_vbox.pack_start(label, expand=False, fill=False)
            label.show()

            hbox = gtk.HBox()
            spacer = gtk.Label(' ' * 4)
            hbox.pack_start(spacer, expand=False, fill=False)
            spacer.show()

            # One row per setting
            vbox = gtk.VBox(spacing=6)

            for i in range(len(groupSettings)):
                settingClass, args = groupSettings[i][0], groupSettings[i][1:]
            
                setting = settingClass(self.config, *args)
                label = setting.createLabel()
                setting.createWidget()
                control = setting.widget
                
                setting.load()
                self.settings.append(setting)

                sizeGroup2.add_widget(control)

                if label:
                    sizeGroup.add_widget(label)
                    label.set_mnemonic_widget(control)
                    setting_hbox = gtk.HBox(spacing=12)
                    setting_hbox.pack_start(label, fill=False)
                    setting_hbox.pack_end(control, expand=False, fill=False)
                    label.show()
                    control.show()
                    setting_hbox.show()
                    vbox.pack_start(setting_hbox, expand=True, fill=True)
                else:
                    vbox.pack_start(control, expand=True, fill=False)
                    control.show()

            # End settings vbox
            hbox.pack_start(vbox, fill=True, expand=False)
            vbox.show()

            group_vbox.pack_start(hbox, expand=False, fill=False)
            hbox.show()

            settingsVBox.pack_start(group_vbox, expand=False, fill=False)
            group_vbox.show()

    def loadConfiguration(self):
        self.config = Configuration()
        self.config.load()

        print self.config.bindings
        print self.config.settings

    def saveConfiguration(self):
        print '*' * 40

        for setting in self.settings:
            try:
                setting.save()
            except ValidationError, e:
                dlg = gtk.MessageDialog(self.get_widget('main_window'),
                                        gtk.DIALOG_DESTROY_WITH_PARENT,
                                        gtk.MESSAGE_ERROR,
                                        gtk.BUTTONS_CLOSE,
                                        e.message)
                dlg.run()
                dlg.destroy()
                e.widget.grab_focus()
                return False
            
        print self.config.bindings
        print self.config.settings

        return True

    def on_button_save_clicked(self, widget):
        if self.saveConfiguration():
            pass
#            gtk.mainquit()
#            self.config.save()
        
    def quit_app(self, widget, data=None):        
        gtk.mainquit()
#        self.config.save()

    # Buttons
    def on_button_add_clicked(self, widget):
        BindingDialog(self,
                      self.get_widget('binding_dialog'),
                      self.get_widget('main_window'),
                      self.currentContextTypes())
    
    # Contexts
    def on_focus_optionmenu_changed(self, widget):
        models = ['click', 'sloppy', 'mouse'] # Same order as in option menu
        
        focusModel = models[widget.get_history()]

        self.config.settings['focus'] = focusModel

    def checkboxHandler(self, widget, settingName):
        if widget.get_active():
            self.config.settings[settingName] = 'true'
        else:
            self.config.settings[settingName] = 'false'

    def comboHandler(self, widget, settingName):
        self.config.settings[settingName] = widget.get_text()
            
    def on_clicktoraise_check_toggled(self, widget):
        self.checkboxHandler(widget, 'clickToRaise')

    def on_focusnew_check_toggled(self, widget):
        self.checkboxHandler(widget, 'focusNewWindows')
        
    def on_contextsmenu_activate(self, widget, context):
        print "Context %s activated" % context

        self.changeToContext(context)

    def refreshActionList(self):
        model = self.bindingsList.get_model()
        model.clear()

        bindings = self.config.bindings.get(self.currentContext, [])
        
        for eventType, params, action in bindings:
            if eventType in ('KeyPress', 'KeyRelease'):
                eventType = '%s (%s %s)' % (eventType,
                                            ' '.join(params['modifier']),
                                            params['name'])
                
            model.append((eventType, action))

    def changeToContext(self, context):
        self.currentContext = context
        self.refreshActionList()

    def appendBinding(self, binding):
        self.config.bindings[self.currentContext].append(binding)
            
    def fillContextsMenu(self):
        contextsMenu = self.get_widget('context_optionmenu')

        menu = gtk.Menu()
        
        for label, value in self.contexts():
            if label is None or value is None:
                menu_item = gtk.SeparatorMenuItem()
            else:
                menu_item = gtk.MenuItem(label)
                menu_item.connect('activate',
                                  self.on_contextsmenu_activate, value)
                
            menu_item.show()
            menu.append(menu_item)

        contextsMenu.set_menu(menu)

    def currentContextTypes(self):
        if self.currentContext.startswith('window.') or \
               self.currentContext == 'defaultAllWindows':
            return ['window', 'screen']
        elif self.currentContext == 'root' or \
                 self.currentContext.startswith('edge.'):
            return ['screen']

    def contexts(self):
        return [
            # Window contexts
            ('All window parts', 'defaultAllWindows'),
            ('Window frame', 'window.frame'),
            ('Client area (Focused)', 'window.activeclient'),
            ('Client area (Unfocused)', 'window.passiveclient'),
            ('Title bar', 'window.title'),
            ('Label (title text)', 'window.label'),
            ('Left resize grip', 'window.leftgrip'),
            ('Right resize grip', 'window.rightgrip'),
            ('Handle bar (bottom)', 'window.handle'),
            # (None, None) means separator
            (None, None),
            # Screen
            ('Root window (desktop)', 'root'),
            # Edges
            ('North (upper) edge', 'edge.north'),
            ('South (lower) edge', 'edge.south'),
            ('West  (left)  edge', 'edge.west'),
            ('East  (right) edge', 'edge.east'),
            ]

def run(file):
	MainWindow(file)
	gtk.main()

# vim: sw=4 ts=4
